/* GENERAL DESCRIPTION

This program can be used to detect whether a given Language E domain D 
WITH RAMIFICATION STATEMENTS as well as t-propositions
entails a given Goal and whether the Goal holds in some model of the domain D.

Goal is a list of holds(F,T) and neg(holds(F,T)), for ground F and T, and possibly true.

The program relies upon a logic programming-based formulation of D (see test files)
and upon an argumentation-based reformulation of entailment and models, by which:

   -  D entails Goal  iff Goal is a sceptical non-monotonic consequence of D

   -  Goals holds in a model of D iff Goal is a credulous non-monotonic consequence of D.

To detect whether Goal is a sceptical non-monotonic consequence
   | ?-sceptical(Goal).

To detect whether Goal is a credulous non-monotonic consequence
   | ?-credulous(Goal).

To detect whether Goal is a credulous non-monotonic consequence and 
to return the corresponding admissible assumption set Delta
   | ?-credulous(Goal,Delta).

Delta is a list of assumption and generation rules, i.e.
rule(ass,F,T), rule(ass,neg(F),T),  rule(gen,F,T,NT),  rule(gen,neg(F),T,NT),
for ground F and T and NT

*/

:- compile('utils').
:- compile('effects').
:- compile('preprocessing').
:- compile('constraints').


%%%%%%%%%%%%%% Top-level predicates %%%%%%%%%%%%%%

sceptical(Goal):- 
	credulous(Goal), !, % TONY 14/10/00
      show(['proved-cred(', Goal, ')']),
	contrarylist(Goal,NGoal),
	\+ credulousone(NGoal).

credulousone(List):- mymember(Goal,List), credulous([Goal]).

credulous(Goal):-
	credulous(Goal,_).

credulous(Goal,Delta0):-
	query_pre_processing(Goal, Relevant),
	show('\n\n\n>> ====================\n>> Computing Delta0...\n>> ====================\n\n\n'),
	relevant(RelevantTprops,_,_,_,_)=Relevant,abd([],RelevantTprops,DeltaKnow),abd(DeltaKnow,Goal,Delta0),
%	abd([],Goal,Delta0),
	show('\n>> Delta0 list:\n'),
	showDelta(Delta0),
	once(validate_constraints(Delta0, Relevant, _)).


%%%%%%%%%%%%%% Middle-level predicates %%%%%%%%%%%%%%

% Addition to allow us to write ramifications with variables.
% It assumes that all variales are typed in the ramification and
% that these are written first in the conditions of the ramification.
% Once we ground all the type predicates the the whole ramification
% should be ground.


% 1.10.01 - Allows for constant fluents (positive and negative) and system predicates
%           to appear anywhere, given that they are properly instantiated at some point.
ground(List, GroundedList):-
	order_list(List, [], [], [], [], OrderedList),
	ground_aux(OrderedList, GroundedList).

order_list([], PosConst, NegConst, System, Rest, OrderedList):-
	list_append([PosConst, NegConst, System, Rest], OrderedList).
order_list([X|List], PosConst, NegConst, System, Rest, OrderedList):-
	X=holds(F,_), constantfluent(F), !,
	order_list(List, [X|PosConst], NegConst, System, Rest, OrderedList).
order_list([X|List], PosConst, NegConst, System, Rest, OrderedList):-
	X=neg(holds(F,_)), constantfluent(F), !,
	order_list(List, PosConst, [X|NegConst], System, Rest, OrderedList).
order_list([X|List], PosConst, NegConst, System, Rest, OrderedList):-
	X=holds(F,_), system_pred(F), !,
	order_list(List, PosConst, NegConst, [X|System], Rest, OrderedList).
order_list([X|List], PosConst, NegConst, System, Rest, OrderedList):-
	order_list(List, PosConst, NegConst, System, [X|Rest], OrderedList).


ground_aux([],[]).

% To handle system predicates - 13.04.01
ground_aux([X|Rest], NewRest):- 
    (X=holds(F,_);X=neg(holds(F,_))),
    system_pred(F), !,
    F,
    ground_aux(Rest, NewRest).   

ground_aux([X|Rest], [X|Rest]):- 
    (X=holds(F,_);X=neg(holds(F,_))),
    \+ nonground(F),
    \+ constantfluent(F), !.              % Assume that the rest is ground

ground_aux([X|Rest], NewRest):- 
    X=holds(F,_),                % 1.10.01 - Positive constant fluents produce grounded clauses.
    constantfluent(F),
    unifyavaluefromtime0(X),
    ground_aux(Rest, NewRest).   % We drop X because this is a constant fluent 

ground_aux([X|Rest], NewRest):- 
    X=neg(holds(F,_)),           % 1.10.01 - Negative constant fluents reduce grounded clauses.
    constantfluent(F),
    \+ unifyavaluefromtime0(holds(F,_)),
    ground_aux(Rest, NewRest).   % We drop X because this is a constant fluent 


% RELEVANT RAMIFICATIONS-Start


validateRelRams(Delta,Delta1):-
   findall(T,clause(happens(_,T),true), Times),
   addzero(Times,TimesZero),
   removeduplicates(TimesZero, CTimes),
   order(CTimes, OrdTimes),
   checkRelevantRams(Delta, OrdTimes, Delta1).

checkRelevantRams(Delta, [], Delta).

checkRelevantRams(Delta,[T|RestTimes],Delta1):-
   nexttimeZero(T,NT),
   checkRelevantRamsAtT(Delta, Delta, NT, Delta2),
   checkRelevantRams(Delta2,RestTimes,Delta1).

nexttimeZero(0,0):-!.
nexttimeZero(T,NT):- T>0, NT is T + 1.

%checkRelevantRamsAtT(Delta, DeltaNew, AtT, DeltaOut)

checkRelevantRamsAtT(Delta, [], AtT, Delta):-!.

checkRelevantRamsAtT(Delta, DeltaNew, AtT, DeltaOut):-
   projectassumptions(Delta, DeltaNew, AtT, LiteralsAtT, DeltaT),
   removeduplicates(LiteralsAtT,CLiteralsAtT), writeln(CLiteralsAtT),
   findRELRAMS(CLiteralsAtT, AtT, Rams),  % Rams here are already resolved.
   genuineappend(DeltaT, Delta, DeltaExp), writelist(Rams),
   satisfyRelRams(DeltaExp, Rams, DeltaInt),
   myappend(DeltaDiff, DeltaExp, DeltaInt), 
   writeln(next), nl,
   checkRelevantRamsAtT(DeltaInt, DeltaDiff, AtT, DeltaOut).

genuineappend([], Delta, Delta).
genuineappend([X|Rest], Delta, RestDelta):- 
   mymember(X, Delta), !, 
   genuineappend(Rest, Delta, RestDelta).

genuineappend([X|Rest], Delta, [X|RestDelta]):-   % \+ mymember(X, Delta)
   genuineappend(Rest, Delta, RestDelta).


satisfyRelRams(Delta, [], Delta).

satisfyRelRams(DeltaExp, [First|Rest], DeltaOut):-
   satisfyRelRamONE(DeltaExp, First, Delta1),
   satisfyRelRams(Delta1, Rest, DeltaOut).

satisfyRelRamONE(DeltaExp, ramification(Head,[]), Delta1):- !,
   abd(DeltaExp, [Head], Delta1).

satisfyRelRamONE(DeltaExp, ramification(false,[X]), Delta1):-
   contrary(X,Y),
   abd(DeltaExp, [Y], Delta1).

satisfyRelRamONE(DeltaExp, ramification(Head,_), Delta1):-
   abd(DeltaExp, [Head], Delta1), 
   contrary(Head,CHead),
   \+ abd(DeltaExp, [CHead], _).

satisfyRelRamONE(DeltaExp, ramification(_,ListConds), Delta1):-
   mymember(X,ListConds),
   contrary(X,Y),
   abd(DeltaExp, [Y], Delta1),
   \+ abd(DeltaExp, [X], _).

satisfyRelRamONE(DeltaExp, ramification(Head,_), Delta1):-
   abd(DeltaExp, [Head], Delta1).

satisfyRelRamONE(DeltaExp, ramification(_,ListConds), Delta1):-
   mymember(X,ListConds),
   contrary(X,Y),
   abd(DeltaExp, [Y], Delta1).

satisfyRelRamONE(DeltaExp, Ram, DeltaExp):-
   writeln(failedRAM(Ram)).


%findRELRAMS(DeltaT, AtT, Rams)

findRELRAMS([], T, []).

findRELRAMS([Literal|RestLits], T, Rams):-
   findRELRAMSOne(Literal, T, RamsOne),
   findRELRAMS(RestLits, T, RestRams),
   myappend(RamsOne, RestRams, MixedRams),
   putsimplefirst(MixedRams, Rams).

putsimplefirst([], []).

putsimplefirst([One|MixedRams], [One|Rams]):-
   (One=ramification(_,[]); One=ramification(false,[_])), !, %Single body. RED CUT
   putsimplefirst(MixedRams, Rams).

putsimplefirst([One|MixedRams], FirstRams):-
   putsimplefirst(MixedRams, Rams),
   myappend(Rams, [One], FirstRams).


findRELRAMSOne(Literal, T, Rams):-
   findall(ramification(ResHead,GroundListRest),
           (clause(ram(Y),Rest),convert(Rest,ListRest),
            resolve(Literal, Y, ResHead, ListRest, RestRest), ground(RestRest,GroundListRest)), 
            Rams).

resolve(Literal, HeadOfRam, false, List, List):- contrary(Literal, HeadOfRam).
resolve(Literal, Y, Y, List, ResolvedList):- myselect(Literal, List, ResolvedList).


%projectassumptions(Delta, Delta, T, DeltaT)
  
projectassumptions(Delta, [], T, [], []).

projectassumptions(Delta, [X|Rest], T, [XatT|LiteralsT], DeltaT):-
   projectONE(Delta, X,T,XatT, NewArgAtT), !,                     % RED CUT.
   projectassumptions(Delta, Rest, T, LiteralsT, DeltaRestT),
   myappend(NewArgAtT, DeltaRestT, DeltaT).
   
projectassumptions(Delta, [X|Rest], T, LiteralsT, DeltaT):-
   projectassumptions(Delta, Rest, T, LiteralsT, DeltaT).

projectONE(Delta,X,T,XatT, NewArgAtT):-
   \+ constantargument(X),
   transformforprojection(X,T,XatT), % INCOMPLETE ON THE 25.11.00
   holdsinDelta(Delta, XatT, NewArgAtT).

holdsinDelta(Delta, XatT, []):- provable(XatT, Delta), !.

holdsinDelta(Delta, XatT,[NewArgAtT]):-
   abd(Delta, [XatT], DeltaNew),!,
   contrary(XatT, Y), 
  \+ abd(Delta, [Y], _), 
  myappend(_, [NewArgAtT|Delta], DeltaNew).   

/* holdsinDelta(Delta, XatT, [NewArgAtT]):-
   abd(Delta, [XatT], [NewArgAtT|Delta]).   
%Here we allow projection only if
%no further arguments are required. May want to replace with:
% abd(Delta, Literal, DeltaNew), \+ abd(Delta, NegLiteral, _). 
% AND MAYBE RECURSE ON DeltaNew. */

constantargument(X):- 
   (X=rule(ass,F,T1); X=rule(ass,neg(F),T1)), 
   constantfluent(F).
                     

constantargument(X):- 
   (X=rule(gen,F,T_,_); X=rule(gen,neg(F),_,_)), 
   constantfluent(F).

%transformforprojection(Argument, Time, LiteralAtTime)
%On the 25.11.00 this is INCOMPLETE.
%Need to complete by allowing reasoning forward with
%c-propositions ONLY? and given h-propositions (before
%time Time) to new possible conslusions under the 
%current Delta. Then RECURSE on this.
transformforprojection(rule(ass,F,T1), T, holds(F,T)):- \+ (F=neg(_)), !.

transformforprojection(rule(ass,neg(F),T1), T, neg(holds(F,T))):- !.

transformforprojection(rule(gen,F,Tar, T1), T, holds(F,T)):- \+ (F=neg(_)), !, T1 =< T.

transformforprojection(rule(gen,neg(F),Tar, T1), T, neg(holds(F,T))):- !, T1 =< T.

% RELEVANT RAMIFICATIONS-End


%%%%%%%%%%

abd(Delta,[],Delta):-!.

abd(Delta,[true],Delta):-!.

%Addition of system Predicates on the 12.04.01
abd(Delta,[X],Delta):- 
   (X=holds(F,_);X=neg(holds(F,_))),
   system_pred(F), !,
   F.

% Addition on the 7.11.00 to handle auxiliary constant predicates
% whose value is the same from time 0 onwards.
abd(Delta,[X],Delta):- \+ nonground(X),  %7.11.00 For auxiliary constant predicates
   provable(X,Delta),!.  

abd(Delta,[X],Delta1):- nonground(X),!,  %7.11.00 Then X must be a constant predicate
   unifyavaluefromtime0(X),            % grounds X from t-propositions at time 0.
   abd(Delta,[X],Delta1).

% End of change at 7.11.00

abd(Delta,[X],Delta1):- \+ nonground(X), !,     
   contrary(X,Y),
   \+ provable(Y,Delta), % consistency check
   subabd(Delta,X,Delta1).

abd(Delta,[X|Y],Delta1):-
   abd(Delta,[X],Deltastar),
   abd(Deltastar,Y,Delta1).


%For change on the 7.11.00 for auxiliary constant predicates.
unifyavaluefromtime0(holds(F,T)):-
   clause(tprop(holds(F,0)),true). %Unifies the variables in F with the values at 0
unifyavaluefromtime0(neg(holds(F,T))):-   
   clause(tprop(neg(holds(F,0))),true). 

subabd(Delta,X,Delta1):-        	%proof by generation
   provablegen(X,Delta,Deltaplus,Precon,STime), 
   consgen(Deltaplus,X,STime,Delta2),
   abd(Delta2,Precon,Delta1).

subabd(Delta,X,Delta1):-            %proof by assumption
   appendass(X,Delta,Deltaplus), 
   consass(Deltaplus,X,Delta1).


%%%%defending assumptions

appendass(holds(F,T),Delta,Deltaplus):-!,
   myappend([rule(ass,F,T)],Delta,Deltaplus). 

appendass(neg(holds(F,T)),Delta,Deltaplus):-!,
   myappend([rule(ass,neg(F),T)],Delta,Deltaplus).

consass(Deltaplus,X,Delta1):- 
   conspreviousass(Deltaplus,Deltaplus,X,Delta2), % Checks persistence attacks on existing assumption rules in Delta.
   conscurrentass(Delta2, X, Delta1).
   
conspreviousass(Delta,[],_,Delta).

conspreviousass(Delta,[Y|Rest],X,Delta2):-
   conspreviousassONE(Delta,Y,X,Delta3), 
   conspreviousass(Delta3,Rest,X,Delta2).
   
conspreviousassONE(Delta,rule(gen,_,_,_),_,Delta):-!.

conspreviousassONE(Delta,rule(ass,F1,_),holds(_,_),Delta):- \+ F1=neg(_),!.

conspreviousassONE(Delta,rule(ass,neg(_),_),neg(holds(_,_)),Delta):-!.

conspreviousassONE(Delta,rule(ass,F,_),neg(holds(F1,_)),Delta):- \+ F=F1,!. 

conspreviousassONE(Delta,rule(ass,F,T1),neg(holds(F,T)),Delta):- T > T1,!. 

conspreviousassONE(Delta,rule(ass,F,T1),neg(holds(F,T)),Delta1):-!, 
   T < T1,
   defenceagainstpers(Delta,holds(F,T1),T, Delta1).

conspreviousassONE(Delta,rule(ass,neg(F),_),holds(F1,_),Delta):- \+ F=F1,!. 
   
conspreviousassONE(Delta,rule(ass,neg(F),T1), holds(F,T),Delta):- T > T1,!. 

conspreviousassONE(Delta,rule(ass,neg(F),T1),holds(F,T),Delta1):-!, 
   T < T1,
   defenceagainstpers(Delta,neg(holds(F,T1)),T, Delta1).

defenceagainstpers(Delta,X,T, Delta1):-   %Defend only by generation.
   provablegen(X,Delta,Delta2,Precon,STime),
   STime > T,     
   consgen(Delta2,X,STime,Delta3),
   abd(Delta3,Precon,Delta1).
   
   
conscurrentass(Delta, X, Delta1):-
   findattackspers(X,Delta, ListofPreviousTimes), %Persistence attacks
   defenceagainstpersMANY(Delta, X, ListofPreviousTimes, Delta2),
   findattacksgen(X,GenAttacks),           %Generation attacks
   defencegen(Delta2,GenAttacks,Delta1).
   

findattackspers(holds(F,T),Delta, ListofPreviousTimes):-
   findall(T1, 
           (mymember(rule(ass,neg(F),T1),Delta), T1 < T), 
           ListofPreviousTimes).
   
findattackspers(neg(holds(F,T)),Delta, ListofPreviousTimes):-
   findall(T1, 
           (mymember(rule(ass,F,T1),Delta), T1 < T), 
           ListofPreviousTimes).
   
defenceagainstpersMANY(Delta, _, [], Delta).   
   
defenceagainstpersMANY(Delta, X, [T|Rest], Delta1):-
   defenceagainstpers(Delta,X,T, Delta2),
   defenceagainstpersMANY(Delta2, X, Rest, Delta1).


findattacksgen(holds(F,T),GenAttacks):-
  findall((PreconList,rule(gen,neg(F),T,NT)),
          (terminated(F,NT,T,PreconList, [neg(holds(F,T))]),NT<T),
          GenAttacks).

findattacksgen(neg(holds(F,T)),GenAttacks):-
  findall((PreconList,rule(gen,F,T,NT)),
          (initiated(F,NT,T,PreconList, [holds(F,T)]),NT<T),
          GenAttacks).

%%%%defending generation rules

provablegen(holds(F,T),Delta,Deltaplus,Precon,NT):-
   initiated(F,NT,T,Precon, [holds(F,T)]),
   myappend([rule(gen,F,T,NT)],Delta,Deltaplus).

provablegen(neg(holds(F,T)),Delta,Deltaplus,Precon,NT):-
   terminated(F,NT,T,Precon, [neg(holds(F,T))]),
   myappend([rule(gen,neg(F),T,NT)],Delta,Deltaplus).


initiated(F,NT,T,Precon, _):-
   get_action_effect_info(holds(F,NT), _, ActPrecon, Conds),
   NT < T,
   nexttimeconds(Conds,CondsNext),
   myappend(ActPrecon,CondsNext,Precon).

terminated(F,NT,T,Precon, _):-
   get_action_effect_info(neg(holds(F,NT)), _, ActPrecon, Conds),
   NT < T,
   nexttimeconds(Conds,CondsNext),
   myappend(ActPrecon,CondsNext,Precon).


started(holds(F,NT),NT,T,Precon, Acc):-
   initiated(F,NT,T,Precon, Acc).

started(neg(holds(F,NT)),NT,T,Precon, Acc):-
   terminated(F,NT,T,Precon, Acc).

membersp(X,Y):- mymember(X,Y). % After fixing the loop checker 18.11.00

membersp(holds(H,_), List):- mymember(H , List).		% Singleton variable T replaced by _
membersp(neg(holds(H,_)), List):- mymember(H , List).	% Singleton variable T replaced by _

consgen(Delta,neg(holds(F,Target)),STime,Delta1):-  %attack by generation
  findall((PreconList,rule(gen,F,Target,NT)),
          (initiated(F,NT,Target,PreconList, [holds(F,Target)]), NT<Target,STime<NT),
          S),
  defencegen(Delta,S,Delta1). 

consgen(Delta,holds(F,Target),STime,Delta1):-
  findall((PreconList,rule(gen,neg(F),Target,NT)),
          (terminated(F,NT,Target,PreconList, [neg(holds(F,Target))]),
           NT<Target,STime<NT),
          S),
  defencegen(Delta,S,Delta1).

defencegen(Delta,[], Delta):-!.

defencegen(Delta,[X|Rest], Delta1):-
  defencegenONE(Delta,X,Delta2),
  defencegen(Delta2,Rest, Delta1).

defencegenONE(Delta,(_,rule(gen,X,TargetTime,NT)), Delta1):- 
   contraryfluent(X,Y,TargetTime),
   provablegen(Y,Delta,Deltaplus,Precon,STime),
   STime > NT,               
   consgen(Deltaplus,Y,STime,Delta2), 
   abd(Delta2,Precon,Delta1).  

defencegenONE(Delta,(Precs,rule(gen,_,_,_)), Delta1):-
   mymember(Prec, Precs),
   contrary(Prec, NPrec),
   abd(Delta, [NPrec], Delta1).
   

%%%%%%%%%%%% Subsidiary predicates %%%%%%%%%%%%

myselect(X,List):- mymember(X,List).

% order(X,Y):-myappend([0],X,Y). 
% insertion sort
order([], []).
order([X|Xs], Ys):-
	order(Xs, Zs),
	insert(X, Zs, Ys).
insert(X, [], [X]).
insert(X, [Y|Ys], [Y|Zs]):- X>Y, insert(X, Ys, Zs).
insert(X, [Y|Ys], [X,Y|Ys]):- X=<Y.

   
settime(NT,holds(F,_),holds(F,NT)).
%settime(NT,holds(_,NT)).
settime(NT,neg(holds(F,_)),neg(holds(F,NT))).

nexttime(T,NT):- NT is T+1.
 
nexttimeconds([],[]).
nexttimeconds([holds(F,T)|Rest],[holds(F,NextT)|NextRest]):-
   nexttime(T,NextT),
   nexttimeconds(Rest,NextRest).
nexttimeconds([neg(holds(F,T))|Rest],[neg(holds(F,NextT))|NextRest]):-
   nexttime(T,NextT),
   nexttimeconds(Rest,NextRest).

myselect(X,[X|Rest],Rest). 
myselect(X,[Y|Rest],[Y|NewRest]):- myselect(X,Rest,NewRest).


contrarylist([Goal],[NGoal]):- 
    contrary(Goal,NGoal).
contrarylist([Goal|Rest], [NGoal|NewRest]):- 
    contrary(Goal,NGoal), 
    contrarylist(Rest, NewRest).

contrary(holds(F,T),neg(holds(F,T))):-!.
contrary(neg(holds(F,T)),holds(F,T)):-!.

contraryfluent(F, neg(holds(F,TargetTime)), TargetTime):- \+ (F = neg(_)).
contraryfluent(neg(F),holds(F,TargetTime), TargetTime).

% Changed the ordering.
provable(neg(holds(F,T)),Delta):-
   mymember(rule(ass,neg(F),T),Delta),!.
provable(holds(F,T),Delta):-   
   mymember(rule(ass,F,T),Delta),!.
provable(neg(holds(F,T)),Delta):-
   mymember(rule(gen,neg(F),T,_),Delta),!.
provable(holds(F,T),Delta):-   
   mymember(rule(gen,F,T,_),Delta),!.

convert((X,Y),[X|Z]) :- !, convert(Y,Z).
convert(true,[]):- !.
convert(X,[X]).

writelist([]).
writelist([X|Rest]):- writeln(X), writelist(Rest).

mymember(X,[X|_]).
mymember(X,[_|Y]):- mymember(X,Y).

myappend([],Y,Y):-!.
myappend(X, [],X):-!.
myappend([X|Y],Z,[X|W]):-myappend(Y,Z,W).

removeduplicates([],[]).
removeduplicates([L|Rest], CList):-
   mymember(L,Rest), !, 
   removeduplicates(Rest, CList).
removeduplicates([L|Rest], [L|CList]):-
   removeduplicates(Rest, CList).

addzero(X,Y):- myappend([0],X,Y).

%SYSTEM PREDICATES

system_pred(A) :-
	functor(A, Functor, Arity),
	current_built_in(Functor/Arity).
