% =============================================================================================
% Loads and binds the "C" functions.
% =============================================================================================

:- load('convert.dll').
:- external(run_sat/1, 'p_run_sat').

% =============================================================================================
% Simulates the backtracking of "satisfiable".
% =============================================================================================

checkallrams(Delta, relevant(RelevantTprops, RelevantRams, _, RelevantLiterals, RelevantTimes), Delta1):-
	checkallrams(Delta, (RelevantRams, RelevantTprops, RelevantLiterals), RelevantTimes, Delta1), !,
	delete_intermediate_files.
checkallrams(_, _, _):-
	delete_intermediate_files, fail.


checkallrams(Delta, _RelevantRamsTpropsLiterals, [], Delta).
checkallrams(Delta, RelevantRamsTpropsLiterals, [T|RestTimes], Delta1):-
	checkallramsattime(Delta, T, RelevantRamsTpropsLiterals, RestTimes, 1, Delta1).


run_check(Delta, NT, RelevantRamsTpropsLiterals, RestTimes, _SolNum, Literals, Delta1):-
	show('\n>> Adding literals to Delta list... '),
	abd(Delta, Literals, Delta2), 
	show(['\n>> Success in adding literals to Delta list, for time point ', NT, ' .\n']),
	checkallrams(Delta2, RelevantRamsTpropsLiterals, RestTimes, Delta1).
run_check(Delta, NT, RelevantRamsTpropsLiterals, RestTimes, SolNum, _Literals, Delta1):-
	show(['\n>> Failure in adding literals to Delta list, for time point ', NT, ' .\n']),
	show('>> Backtracking...\n'),
	NextSolNum is SolNum+1, 
	checkallramsattime(Delta, NT, RelevantRamsTpropsLiterals, RestTimes, NextSolNum, Delta1).


% =============================================================================================
% Calls the SAT Solver and returns a solution, given the Ramifications and a Help set.
% =============================================================================================

satisfiable(RamsIn, Help, SolNum, NT, RamsOut):-
	open('rams.in',write,fos),
	write_help_to_file(Help), write_rams_to_file(RamsIn),
	close(fos),
	run_sat(SolNum),
	open('rams.out',read,fis),
	read_from_file(NT, RamsOut),
	close(fis), !.
satisfiable(_, _, _, _, _):-
	show('No solution found. Backtracking...\n'),
	fail.


write_help_to_file([]).
write_help_to_file([Literal|Rest]):-
	write_clause_to_file([Literal]), write_help_to_file(Rest).

write_rams_to_file([]).
write_rams_to_file([ramification(Head,Body)|Rest]):-
	write_clause_to_file([Head|Body]), write_rams_to_file(Rest).

write_clause_to_file([]):- write(fos, '\n'), flush(fos).
write_clause_to_file([holds(F,_T)|Rest]):-
	write(fos, holds(F)), write(fos, ';'), write_clause_to_file(Rest).
write_clause_to_file([neg(holds(F,_T))|Rest]):-
	write(fos, neg(holds(F))), write(fos, ';'), write_clause_to_file(Rest).


read_from_file(NT, [LiteralAtNT|Rest]):- 
	read(fis, Literal),
	Literal \== 'end_of_file',
	add_time(Literal, NT, LiteralAtNT), !, 
	read_from_file(NT, Rest).
read_from_file(_NT,[]).

add_time(holds(F), NT, holds(F, NT)).
add_time(neg(holds(F)), NT, neg(holds(F, NT))).


delete_intermediate_files:-
	delete_files(['rams.in', 'cnf.in', 'cnf.out', 'rams.out']).

delete_files([]).
delete_files([F|Rest]):-
	exists(F), !,
	delete(F),
	delete_files(Rest).
delete_files([_|Rest]):-
	delete_files(Rest).


% =============================================================================================
% Useful predicates for Sat based versions.
% =============================================================================================

showTimePoint(NT, Delta):-
	show(['\n\n>> Time point: ', NT, '\n', '>> ===========\n']),
	show('\n>> Input Delta list:\n'), showDelta(Delta).

project_Delta(Delta, NT, SkepticalLiteralsAtNT):-
	show('\n>> Skeptically projecting Delta at time point...\n'), 
	skeptically_project_Delta_at_Time(Delta, NT, SkepticalLiteralsAtNT),
	show('\n>> Projected literals list:\n'), showList(SkepticalLiteralsAtNT).

project_relevant(Delta, RelevantLiterals, SkepticalRelevantLiterals):-
	show('\n>> Skeptically projecting relevant literals at time point...\n'), 
	find_skeptically_projected_Literals(Delta, RelevantLiterals, SkepticalRelevantLiterals),
	show('\n>> Projected skeptical literals list:\n'), showList(SkepticalRelevantLiterals).

find_Delta_relevant(Delta, RelevantLiterals):-
	show('\n>> Finding relevant literals from Delta... '), 
	find_relevant_literals_from_Delta(Delta, RelevantLiterals, _, _),
	show('\n>> Relevant literals list:\n'), showList(RelevantLiterals).

find_relevant(SeedLiterals, RelevantLiterals):-
	show('\n>> Finding relevant literals... '), 
	find_relevant_literals(SeedLiterals, RelevantLiterals, _, _),
	show('\n>> Relevant literals list:\n'), showList(RelevantLiterals).

filter_solution(SeedLiterals, SatOut, FilteredLiterals):-
	show('\n>> Filtering relevant solution... '), 
	find_relevant_literals(SeedLiterals, RelevantLiterals, _, _),
	filter_Literals(SatOut, RelevantLiterals, FilteredLiterals),
	show('\n>> Filtered solution:\n'), showList(FilteredLiterals).

filter_Literals([], _, []).
filter_Literals([L|AllLiterals], FilterLiterals, [L|KeptLiterals]):-
	mymember(L, FilterLiterals), !,
	filter_Literals(AllLiterals, FilterLiterals, KeptLiterals).
filter_Literals([L|AllLiterals], FilterLiterals, [L|KeptLiterals]):-
	contrary(L, NL), mymember(NL, FilterLiterals), !,
	filter_Literals(AllLiterals, FilterLiterals, KeptLiterals).
filter_Literals([_|AllLiterals], FilterLiterals, KeptLiterals):-
	filter_Literals(AllLiterals, FilterLiterals, KeptLiterals).

% =============================================================================================

call_sat(Rams, Help, SolNum, NT, SatOut):-
	show('\n>> Calling SAT Solver... '),
	satisfiable(Rams, Help, SolNum, NT, SatOut),
	show(['Solution: ', SolNum]),
	show('\n>> Returned solution:\n'), showList(SatOut).

call_sat_with_no_help(Rams, TProps, SolNum, NT, SatOut):-
	find_literals_at_time(TProps, NT, TPropsAtNT),
	call_sat(Rams, TPropsAtNT, SolNum, NT, SatOut).

call_sat_with_minimal_help(Delta, Rams, TProps, SolNum, NT, SatOut, SkepticalLiteralsAtNT):-
	find_literals_at_time(TProps, NT, TPropsAtNT),
	project_Delta(Delta, NT, SkepticalLiteralsAtNT),
	myappend(TPropsAtNT, SkepticalLiteralsAtNT, AllAtNT),
	call_sat(Rams, AllAtNT, SolNum, NT, SatOut).

call_sat_with_extended_help(Delta, Rams, TProps, SolNum, NT, SatOut, SkepticalLiteralsAtNT, RelevantLiterals, SkepticalRelevantLiterals):-
	find_literals_at_time(TProps, NT, TPropsAtNT),
	project_Delta(Delta, NT, SkepticalLiteralsAtNT),
	find_relevant(SkepticalLiteralsAtNT, RelevantLiterals),
	project_relevant(Delta, RelevantLiterals, SkepticalRelevantLiterals),
	myappend(TPropsAtNT, SkepticalRelevantLiterals, AllAtNT),
	call_sat(Rams, AllAtNT, SolNum, NT, SatOut).

call_sat_with_full_help(Delta, Rams, TProps, Literals, SolNum, NT, SatOut, SkepticalRelevantLiterals):-
	find_literals_at_time(TProps, NT, TPropsAtNT),
	set_all_times(Literals, NT, LiteralsNT),
	project_relevant(Delta, LiteralsNT, SkepticalRelevantLiterals),
	myappend(TPropsAtNT, SkepticalRelevantLiterals, AllAtNT),
	call_sat(Rams, AllAtNT, SolNum, NT, SatOut).


% =============================================================================================
% End of file.
% =============================================================================================
